@@ -9,15 +9,13 @@
 #include "mulmod.h"
 #include "cache.h"
 #include "primality.h"
+#include "montmath.h"
 #define FUNC_isqrt  1
 #define FUNC_gcd_ui 1
 #define FUNC_is_perfect_square 1
 #define FUNC_clz 1
 #include "util.h"
 
-/* factor will do trial division through this prime index, must be in table */
-#define TRIAL_TO_PRIME 84
-
 /*
  * You need to remember to use UV for unsigned and IV for signed types that
  * are large enough to hold our data.
@@ -57,30 +55,78 @@ static const unsigned short primes_small[] =
 int factor(UV n, UV *factors)
 {
   int nfactors = 0;           /* Number of factored in factors result */
-  UV f = 7;
+  unsigned int f = 7;
 
   if (n > 1) {
     while ( (n & 1) == 0 ) { factors[nfactors++] = 2; n /= 2; }
     while ( (n % 3) == 0 ) { factors[nfactors++] = 3; n /= 3; }
     while ( (n % 5) == 0 ) { factors[nfactors++] = 5; n /= 5; }
+  }
 
-    if (f*f <= n) {
-      UV sp = 3;
-      while (++sp < TRIAL_TO_PRIME) {
-        f = primes_small[sp];
-        if (f*f > n) break;
+  if (f*f <= n) {
+    UV sp = 4, lastsp = 83;
+    /* Trial division from 7 to 421.  Use 32-bit if possible. */
+    if (n <= 4294967295U) {
+      unsigned int un = n;
+      while (sp < lastsp) {
+        while ( (un%f) == 0 ) {
+          factors[nfactors++] = f;
+          un /= f;
+        }
+        f = primes_small[++sp];
+        if (f*f > un) break;
+      }
+      n = un;
+    } else {
+      while (sp < lastsp) {
         while ( (n%f) == 0 ) {
           factors[nfactors++] = f;
           n /= f;
         }
+        f = primes_small[++sp];
+        if (f*f > n) break;
       }
     }
+    /* If n is small and still composite, finish it here */
+    if (n < 2011*2011 && f*f <= n) {  /* Trial division from 431 to 2003 */
+      unsigned int un = n;
+      while (sp < NPRIMES_SMALL) {
+        while ( (un%f) == 0 ) {
+          factors[nfactors++] = f;
+          un /= f;
+        }
+        f = primes_small[++sp];
+        if (f*f > un) break;
+      }
+      n = un;
+    }
   }
-  if (n < f*f) {
-    if (n != 1)
-      factors[nfactors++] = n;
+  if (f*f > n) {
+    if (n != 1) factors[nfactors++] = n;
     return nfactors;
   }
+#if BITS_PER_WORD == 64   /* this is slower in 32-bit */
+  /* For small values less than f^3, use simple factor to split semiprime */
+  if (n < 200000000 && n < f*f*f) {
+    if (is_prob_prime(n)) factors[nfactors++] = n;
+    else                  nfactors += holf_factor(n, factors+nfactors, 1000000);
+    return nfactors;
+  }
+#endif
+
+  /* Perfect powers.  Factor root only once. */
+  {
+    int i, j, k = powerof(n);
+    if (k > 1) {
+      UV p = rootof(n, k);
+      int nsmallfactors = nfactors;
+      nfactors = factor(p, factors+nsmallfactors);
+      for (i = nfactors; i >= 0; i--)
+        for (j = 0; j < k; j++)
+          factors[nsmallfactors+k*i+j] = factors[nsmallfactors+i];
+      return nsmallfactors + k*nfactors;
+    }
+  }
 
   {
   UV tofac_stack[MPU_MAX_FACTORS+1];
@@ -90,18 +136,28 @@ int factor(UV n, UV *factors)
 
   /* loop over each remaining factor, until ntofac == 0 */
   do {
-    //while ( (n >= f*f) && (!_XS_is_prime(n)) ) {
     while ( (n >= f*f) && (!is_prob_prime(n)) ) {
       int split_success = 0;
-      /* Adjust the number of rounds based on the number size */
-      UV const br_rounds = ((n>>29) < 100000) ?  1500 :  4000;
-      UV const sq_rounds =100000; /* 20k 91%, 40k 98%, 80k 99.9%, 120k 99.99% */
+      /* Adjust the number of rounds based on the number size and speed */
+#if USE_MONTMATH
+      UV const br_rounds = ((n>>29)<100000) ?  8000 : 80000;
+#elif MULMODS_ARE_FAST
+      UV const br_rounds = ((n>>29)<100000) ?   500 :  2000;
+#else
+      UV const br_rounds = (n < (UV_MAX>>2))?     0 : 10000;
+#endif
+      UV const sq_rounds = 200000; /* 20k 91%, 40k 98%, 80k 99.9%, 120k 99.99%*/
 
       /* 99.7% of 32-bit, 94% of 64-bit random inputs factored here */
-      if (!split_success) {
+      if (!split_success && br_rounds > 0) {
         split_success = pbrent_factor(n, tofac_stack+ntofac, br_rounds, 3)-1;
         if (verbose) { if (split_success) printf("pbrent 1:  %"UVuf" %"UVuf"\n", tofac_stack[ntofac], tofac_stack[ntofac+1]); else printf("pbrent 0\n"); }
       }
+      /* Give larger inputs a run with p-1 before SQUFOF */
+      if (!split_success && n > (UV_MAX >> 15) && MULMODS_ARE_FAST) {
+        split_success = pminus1_factor(n, tofac_stack+ntofac, 1000, 15000)-1;
+        if (verbose) printf("small p-1 %d\n", split_success);
+      }
       /* SQUFOF with these parameters gets 99.9% of everything left */
       if (!split_success && n < (UV_MAX>>2)) {
         split_success = squfof_factor(n,tofac_stack+ntofac, sq_rounds)-1;
@@ -155,10 +211,10 @@ int factor(UV n, UV *factors)
 
   /* Sort the non-small factors */
   for (i = nsmallfactors+1; i < nfactors; i++) {
-    UV f = factors[i];
-    for (j = i; j > 0 && factors[j-1] > f; j--)
+    UV fi = factors[i];
+    for (j = i; j > 0 && factors[j-1] > fi; j--)
       factors[j] = factors[j-1];
-    factors[j] = f;
+    factors[j] = fi;
   }
   }
   return nfactors;
@@ -167,19 +223,18 @@ int factor(UV n, UV *factors)
 
 int factor_exp(UV n, UV *factors, UV* exponents)
 {
-  int i, j, nfactors;
+  int i = 1, j = 1, nfactors;
 
   if (n == 1) return 0;
-  /* MPUassert(factors != 0, "factors array is null"); */
   nfactors = factor(n, factors);
 
   if (exponents == 0) {
-    for (i = 1, j = 1; i < nfactors; i++)
+    for (; i < nfactors; i++)
       if (factors[i] != factors[i-1])
         factors[j++] = factors[i];
   } else {
     exponents[0] = 1;
-    for (i = 1, j = 1; i < nfactors; i++) {
+    for (; i < nfactors; i++) {
       if (factors[i] != factors[i-1]) {
         exponents[j] = 1;
         factors[j++] = factors[i];
@@ -244,35 +299,29 @@ int trial_factor(UV n, UV *factors, UV maxtrial)
 }
 
 
-static int _divisors_from_factors(UV v, UV npe, UV* fp, UV* fe, UV* res) {
-  UV p, e, i;
-  if (npe == 0) return 0;
-  p = *fp++;
-  e = *fe++;
-  if (npe == 1) {
-    for (i = 0; i <= e; i++) {
-      *res++ = v;
-      v *= p;
+static void _divisors_from_factors(UV nfactors, UV* fp, UV* fe, UV* res) {
+  UV s, count = 1;
+
+  res[0] = 1;
+  for (s = 0; s < nfactors; s++) {
+    UV i, j, scount = count, p = fp[s], e = fe[s], mult = 1;
+    for (j = 0; j < e; j++) {
+      mult *= p;
+      for (i = 0; i < scount; i++)
+        res[count++] = res[i] * mult;
     }
-    return e+1;
-  } else {
-    int nret = 0;;
-    for (i = 0; i <= e; i++) {
-      int nres = _divisors_from_factors(v, npe-1, fp, fe, res);
-      v *= p;
-      res += nres;
-      nret += nres;
-    }
-    return nret;
   }
 }
 
+static int numcmp(const void *a, const void *b)
+  { const UV *x = a, *y = b; return (*x > *y) ? 1 : (*x < *y) ? -1 : 0; }
+
 UV* _divisor_list(UV n, UV *num_divisors)
 {
   UV factors[MPU_MAX_FACTORS+1];
   UV exponents[MPU_MAX_FACTORS+1];
   UV* divs;
-  int i, j, nfactors, ndivisors;
+  int i, nfactors, ndivisors;
 
   if (n <= 1) {
     New(0, divs, 2, UV);
@@ -287,19 +336,10 @@ UV* _divisor_list(UV n, UV *num_divisors)
   for (i = 1; i < nfactors; i++)
     ndivisors *= (exponents[i] + 1);
   New(0, divs, ndivisors, UV);
-  (void) _divisors_from_factors(1, nfactors, factors, exponents, divs);
-  { /* Sort (Shell sort is easy and efficient) */
-    static int gaps[] = {301, 132, 57, 23, 10, 4, 1, 0};
-    int gap, gapi = 0;
-    for (gap = gaps[gapi]; gap > 0; gap = gaps[++gapi]) {
-      for (i = gap; i < ndivisors; i++) {
-        UV v = divs[i];
-        for (j = i; j >= gap && divs[j-gap] > v; j -= gap)
-          divs[j] = divs[j-gap];
-        divs[j] = v;
-      }
-    }
-  }
+  _divisors_from_factors(nfactors, factors, exponents, divs);
+  /* Sort divisors (numeric ascending) */
+  qsort(divs, ndivisors, sizeof(UV), numcmp);
+  /* Return number of divisors and list */
   *num_divisors = ndivisors;
   return divs;
 }
@@ -311,11 +351,12 @@ UV* _divisor_list(UV n, UV *num_divisors)
  *    1 + p^k + p^k^2 + ... p^k^e
  * Return 0 if the result overflowed.
  */
-static const UV sigma_overflow[5] =
+static const UV sigma_overflow[11] =
 #if BITS_PER_WORD == 64
-         {UVCONST(3000000000000000000),UVCONST(3000000000),2487240,64260,7026};
+         {UVCONST(3000000000000000000),UVCONST(3000000000),2487240,64260,7026,
+         1622, 566, 256, 139, 85, 57};
 #else
-         {UVCONST(          845404560),             52560,    1548,  252,  84};
+         {UVCONST(845404560), 52560, 1548, 252, 84, 41, 24, 16, 12, 10, 8};
 #endif
 UV divisor_sum(UV n, UV k)
 {
@@ -323,7 +364,7 @@ UV divisor_sum(UV n, UV k)
   int nfac, i, j;
   UV product = 1;
 
-  if (k > 5 || (k > 0 && n >= sigma_overflow[k-1])) return 0;
+  if (k > 11 || (k > 0 && n >= sigma_overflow[k-1])) return 0;
   if (n <= 1)                               /* n=0  divisors are [0,1] */
     return (n == 1) ? 1 : (k == 0) ? 2 : 1; /* n=1  divisors are [1]   */
   nfac = factor(n,factors);
@@ -365,6 +406,20 @@ UV divisor_sum(UV n, UV k)
 
 
 
+static int found_factor(UV n, UV f, UV* factors)
+{
+  UV f2 = n/f;
+  int i = f > f2;
+  if (f == 1 || f2 == 1) {
+    factors[0] = n;
+    return 1;
+  }
+  factors[i] = f;
+  factors[1-i] = f2;
+  MPUassert( factors[0] * factors[1] == n , "incorrect factoring");
+  return 2;
+}
+
 /* Knuth volume 2, algorithm C.
  * Very fast for small numbers, grows rapidly.
  * SQUFOF is better for numbers nearing the 64-bit limit.
@@ -388,25 +443,35 @@ int fermat_factor(UV n, UV *factors, UV rounds)
     } while (r > 0);
   }
   r = (x-y)/2;
-  if ( (r != 1) && ((UV)r != n) ) {
-    factors[0] = r;
-    factors[1] = n/r;
-    MPUassert( factors[0] * factors[1] == n , "incorrect factoring");
-    return 2;
-  }
-  factors[0] = n;
-  return 1;
+  return found_factor(n, r, factors);
 }
 
-/* Hart's One Line Factorization.
- * Missing premult (hard to do in native precision without overflow)
- */
+/* Hart's One Line Factorization. */
 int holf_factor(UV n, UV *factors, UV rounds)
 {
   UV i, s, m, f;
 
   MPUassert( (n >= 3) && ((n%2) != 0) , "bad n in holf_factor");
 
+  if (n <= (UV_MAX >> 6)) {    /* Try with premultiplier first */
+    UV npre = n * ( (n <= (UV_MAX >> 13)) ? 720 :
+                    (n <= (UV_MAX >> 11)) ? 480 :
+                    (n <= (UV_MAX >> 10)) ? 360 :
+                    (n <= (UV_MAX >>  8)) ?  60 : 30 );
+    UV ni = npre;
+    while (rounds--) {
+      s = isqrt(ni) + 1;
+      m = (s*s) - ni;
+      if (is_perfect_square(m)) {
+        f = gcd_ui(n, s - isqrt(m));
+        if (f > 1 && f < n)
+          return found_factor(n, f, factors);
+      }
+      if (ni >= (ni+npre)) break;
+      ni += npre;
+    }
+  }
+
   for (i = 1; i <= rounds; i++) {
     s = (UV) sqrt( (double)n * (double)i );
     /* Assume s^2 isn't a perfect square.  We're rapidly losing precision
@@ -417,12 +482,7 @@ int holf_factor(UV n, UV *factors, UV rounds)
       f = isqrt(m);
       f = gcd_ui( (s>f) ? s-f : f-s, n);
       /* This should always succeed, but with overflow concerns.... */
-      if ((f == 1) || (f == n))
-        break;
-      factors[0] = f;
-      factors[1] = n/f;
-      MPUassert( factors[0] * factors[1] == n , "incorrect factoring");
-      return 2;
+      return found_factor(n, f, factors);
     }
   }
   factors[0] = n;
@@ -430,20 +490,82 @@ int holf_factor(UV n, UV *factors, UV rounds)
 }
 
 
+#if USE_MONTMATH
+/* Pollard / Brent.  Brent's modifications to Pollard's Rho.  Maybe faster. */
+int pbrent_factor(UV n, UV *factors, UV rounds, UV a)
+{
+  UV f, m, r, Xi, Xm;
+  const UV inner = (n <= 4000000000UL) ? 32 : 160;
+  int fails = 6;
+  const uint64_t npi = mont_inverse(n),  mont1 = mont_get1(n);
+  Xi = Xm = mont_get2(n);
+  a = mont_geta(a,n);
+  MPUassert( (n >= 3) && ((n%2) != 0) , "bad n in pbrent_factor");
+
+  r = 1;
+  f = 1;
+  while (rounds > 0) {
+    UV rleft = (r > rounds) ? rounds : r;
+    UV saveXi = Xi;
+    /* Do rleft rounds, inner at a time */
+    while (rleft > 0) {
+      UV dorounds = (rleft > inner) ? inner : rleft;
+      saveXi = Xi;
+      rleft -= dorounds;
+      rounds -= dorounds;
+      Xi = mont_sqrmod(Xi,n);  Xi = addmod(Xi,a,n);
+      m = (Xi>Xm) ? Xi-Xm : Xm-Xi;
+      while (--dorounds > 0) {         /* Now do inner-1=63 more iterations */
+        Xi = mont_sqrmod(Xi,n);  Xi = addmod(Xi,a,n);
+        f = (Xi>Xm) ? Xi-Xm : Xm-Xi;
+        m = mont_mulmod(m, f, n);
+      }
+      f = gcd_ui(m, n);
+      if (f != 1)
+        break;
+    }
+    /* If f == 1, then we didn't find a factor.  Move on. */
+    if (f == 1) {
+      r *= 2;
+      Xm = Xi;
+      continue;
+    }
+    if (f == n) {  /* back up, with safety */
+      Xi = saveXi;
+      do {
+        Xi = mont_sqrmod(Xi,n);  Xi = addmod(Xi,a,n);
+        f = gcd_ui( (Xi>Xm) ? Xi-Xm : Xm-Xi, n);
+      } while (f == 1 && r-- != 0);
+    }
+    if (f == 0 || f == n) {
+      if (fails-- <= 0) break;
+      Xm = addmod(Xm, 2, n);
+      Xi = Xm;
+      a++;
+      continue;
+    }
+    return found_factor(n, f, factors);
+  }
+  factors[0] = n;
+  return 1;
+}
+#else
 /* Pollard / Brent.  Brent's modifications to Pollard's Rho.  Maybe faster. */
 int pbrent_factor(UV n, UV *factors, UV rounds, UV a)
 {
   UV f, m, r;
   UV Xi = 2;
   UV Xm = 2;
-  const UV inner = 64;
+  const UV inner = (n <= 4000000000UL) ? 32 : 160;
+  int fails = 6;
 
   MPUassert( (n >= 3) && ((n%2) != 0) , "bad n in pbrent_factor");
 
   r = 1;
+  f = 1;
   while (rounds > 0) {
     UV rleft = (r > rounds) ? rounds : r;
-    UV saveXi;
+    UV saveXi = Xi;
     /* Do rleft rounds, inner at a time */
     while (rleft > 0) {
       UV dorounds = (rleft > inner) ? inner : rleft;
@@ -473,16 +595,20 @@ int pbrent_factor(UV n, UV *factors, UV rounds, UV a)
         Xi = sqraddmod(Xi, a, n);
         f = gcd_ui( (Xi>Xm) ? Xi-Xm : Xm-Xi, n);
       } while (f == 1 && r-- != 0);
-      if ( (f == 1) || (f == n) ) break;
     }
-    factors[0] = f;
-    factors[1] = n/f;
-    MPUassert( factors[0] * factors[1] == n , "incorrect factoring");
-    return 2;
+    if (f == 0 || f == n) {
+      if (fails-- <= 0) break;
+      Xm = addmod(Xm, 2, n);
+      Xi = Xm;
+      a++;
+      continue;
+    }
+    return found_factor(n, f, factors);
   }
   factors[0] = n;
   return 1;
 }
+#endif
 
 /* Pollard's Rho. */
 int prho_factor(UV n, UV *factors, UV rounds)
@@ -491,6 +617,7 @@ int prho_factor(UV n, UV *factors, UV rounds)
   const UV inner = 64;
   UV U = 7;
   UV V = 7;
+  int fails = 3;
 
   MPUassert( (n >= 3) && ((n%2) != 0) , "bad n in prho_factor");
 
@@ -526,13 +653,15 @@ int prho_factor(UV n, UV *factors, UV rounds)
         V = sqraddmod(V, a, n);
         f = gcd_ui( (U > V) ? U-V : V-U, n);
       } while (f == 1 && i-- != 0);
-      if ( (f == 1) || (f == n) )
-        break;
     }
-    factors[0] = f;
-    factors[1] = n/f;
-    MPUassert( factors[0] * factors[1] == n , "incorrect factoring");
-    return 2;
+    if (f == 0 || f == n) {
+      if (fails-- <= 0) break;
+      U = addmod(U,2,n);
+      V = U;
+      a++;
+      continue;
+    }
+    return found_factor(n, f, factors);
   }
   factors[0] = n;
   return 1;
@@ -541,49 +670,52 @@ int prho_factor(UV n, UV *factors, UV rounds)
 /* Pollard's P-1 */
 int pminus1_factor(UV n, UV *factors, UV B1, UV B2)
 {
-  UV f;
-  UV q = 2;
-  UV a = 2;
-  UV savea = 2;
-  UV saveq = 2;
+  UV f, k, kmin;
+  UV a     = 2, q     = 2;
+  UV savea = 2, saveq = 2;
   UV j = 1;
   UV sqrtB1 = isqrt(B1);
   MPUassert( (n >= 3) && ((n%2) != 0) , "bad n in pminus1_factor");
 
-  START_DO_FOR_EACH_PRIME(2, sqrtB1) {
-    UV k = p*p;
-    UV kmin = B1/p;
-    while (k <= kmin)
-      k *= p;
-    a = powmod(a, k, n);
-    q = p;
-  } END_DO_FOR_EACH_PRIME
-  if (a == 0) { factors[0] = n; return 1; }
-  f = gcd_ui(a-1, n);
-  if (f == 1) {
-    savea = a;
-    saveq = q;
-    START_DO_FOR_EACH_PRIME(q+1, B1) {
-      q = p;
-      a = powmod(a, q, n);
+  if (B1 <= primes_small[NPRIMES_SMALL-2]) {
+    UV i;
+    for (i = 1; primes_small[i] <= B1; i++) {
+      q = k = primes_small[i];
+      if (q <= sqrtB1) {
+        k = q*q;  kmin = B1/q;
+        while (k <= kmin)  k *= q;
+      }
+      a = powmod(a, k, n);
       if ( (j++ % 32) == 0) {
         if (a == 0 || gcd_ui(a-1, n) != 1)
           break;
-        savea = a;
-        saveq = q;
+        savea = a;  saveq = q;
+      }
+    }
+  } else {
+    START_DO_FOR_EACH_PRIME(2, B1) {
+      q = k = p;
+      if (q <= sqrtB1) {
+        k = q*q;  kmin = B1/q;
+        while (k <= kmin)  k *= q;
+      }
+      a = powmod(a, k, n);
+      if ( (j++ % 32) == 0) {
+        if (a == 0 || gcd_ui(a-1, n) != 1)
+          break;
+        savea = a;  saveq = q;
       }
     } END_DO_FOR_EACH_PRIME
-    if (a == 0) { factors[0] = n; return 1; }
-    f = gcd_ui(a-1, n);
   }
+  if (a == 0) { factors[0] = n; return 1; }
+  f = gcd_ui(a-1, n);
+
   /* If we found more than one factor in stage 1, backup and single step */
   if (f == n) {
     a = savea;
     START_DO_FOR_EACH_PRIME(saveq, B1) {
-      UV k = p;
-      UV kmin = B1/p;
-      while (k <= kmin)
-        k *= p;
+      k = p;  kmin = B1/p;
+      while (k <= kmin)  k *= p;
       a = powmod(a, k, n);
       f = gcd_ui(a-1, n);
       q = p;
@@ -648,14 +780,7 @@ int pminus1_factor(UV n, UV *factors, UV B1, UV B2)
     } END_DO_FOR_EACH_PRIME
     f = gcd_ui(b, n);
   }
-  if ( (f != 1) && (f != n) ) {
-    factors[0] = f;
-    factors[1] = n/f;
-    MPUassert( factors[0] * factors[1] == n , "incorrect factoring");
-    return 2;
-  }
-  factors[0] = n;
-  return 1;
+  return found_factor(n, f, factors);
 }
 
 /* Simple Williams p+1 */
@@ -706,14 +831,7 @@ int pplus1_factor(UV n, UV *factors, UV B1)
     }
   } END_DO_FOR_EACH_PRIME
 
-  if ( (f != 1) && (f != n) ) {
-    factors[0] = f;
-    factors[1] = n/f;
-    MPUassert( factors[0] * factors[1] == n , "incorrect factoring");
-    return 2;
-  }
-  factors[0] = n;
-  return 1;
+  return found_factor(n, f, factors);
 }
 
 
@@ -781,7 +899,7 @@ static UV squfof_unit(UV n, mult_t* mult_save)
       SQUARE_SEARCH_ITERATION;
     }
     S = isqrt(Qn);
-    /* printf("found square %lu after %lu iterations with mult %d\n", Qn, i, mult_save->mult); */
+    /* printf("found square %"UVuf" after %"UVuf" iterations with mult %d\n", Qn, i, mult_save->mult); */
 
     /* Reduce to G0 */
     Ro = P + S*((b0 - P)/S);
@@ -884,12 +1002,8 @@ int squfof_factor(UV n, UV *factors, UV rounds)
       if (f64 > 1) {
         if (f64 != mult) {
           f64 /= gcd_ui(f64, mult);
-          if (f64 != 1) {
-            factors[0] = f64;
-            factors[1] = n / f64;
-            MPUassert( factors[0] * factors[1] == n , "incorrect factoring");
-            return 2;
-          }
+          if (f64 != 1)
+            return found_factor(n, f64, factors);
         }
         /* Found trivial factor.  Quit working with this multiplier. */
         mult_save[i].valid = 0;
@@ -907,17 +1021,46 @@ int squfof_factor(UV n, UV *factors, UV rounds)
   return 1;
 }
 
-UV dlp_trial(UV a, UV g, UV p, UV maxrounds) {
-  UV t, k = 1;
+static UV dlp_trial(UV a, UV g, UV p, UV maxrounds) {
+  UV k, t;
   if (maxrounds > p) maxrounds = p;
-  for (k = 1; k < maxrounds; k++) {
-    t = powmod(g, k, p);
-    if (t == a)
-      return k;
+
+#if USE_MONTMATH
+  if (p&1) {
+    const uint64_t npi = mont_inverse(p),  mont1 = mont_get1(p);
+    g = mont_geta(g, p);
+    a = mont_geta(a, p);
+    for (t = g, k = 1; k < maxrounds; k++) {
+      if (t == a)
+        return k;
+      t = mont_mulmod(t, g, p);
+      if (t == g) break;   /* Stop at cycle */
+    }
+  } else
+#endif
+  {
+    for (t = g, k = 1; k < maxrounds; k++) {
+      if (t == a)
+        return k;
+      t = mulmod(t, g, p);
+      if (t == g) break;   /* Stop at cycle */
+    }
   }
   return 0;
 }
 
+/******************************************************************************/
+/* DLP - Pollard Rho */
+/******************************************************************************/
+
+/* Compare with Pomerance paper (dartmouth dtalk4):
+ * Type I/II/III = our case 1, 0, 2.
+ * x_i = u, a_i = v, b_i = w
+ *
+ * Also see Bai/Brent 2008 for many ideas to speed this up.
+ * https://maths-people.anu.edu.au/~brent/pd/rpb231.pdf
+ * E.g. Teske adding-walk, Brent's cycle algo, Teske modified cycle
+ */
 #define pollard_rho_cycle(u,v,w,p,n,a,g) \
     switch (u % 3) { \
       case 0: u = mulmod(u,u,p);  v = mulmod(v,2,n);  w = mulmod(w,2,n); break;\
@@ -925,37 +1068,416 @@ UV dlp_trial(UV a, UV g, UV p, UV maxrounds) {
       case 2: u = mulmod(u,g,p);                      w = addmod(w,1,n); break;\
     }
 
-UV dlp_prho(UV a, UV g, UV p, UV maxrounds) {
-  UV i;
-  UV n = znorder(g, p);
-  UV u=1, v=0, w=0;
-  UV U=u, V=v, W=w;
-  int const verbose = _XS_get_verbose();
+typedef struct prho_state_t {
+  UV u;
+  UV v;
+  UV w;
+  UV U;
+  UV V;
+  UV W;
+  UV round;
+  int failed;
+  int verbose;
+} prho_state_t;
 
-  if (verbose > 1 && n != p-1) printf("for g=%lu p=%lu, order is %lu\n", g, p, n);
-  if (maxrounds > n) maxrounds = n;
-  for (i = 1; i < maxrounds; i++) {
+static UV dlp_prho_uvw(UV a, UV g, UV p, UV n, UV rounds, prho_state_t *s) {
+  UV i, k = 0;
+  UV u=s->u, v=s->v, w=s->w;
+  UV U=s->U, V=s->V, W=s->W;
+  int const verbose = s->verbose;
+
+  if (s->failed)  return 0;
+  if (s->round + rounds > n)  rounds = n - s->round;
+
+  for (i = 1; i <= rounds; i++) {
     pollard_rho_cycle(u,v,w,p,n,a,g);   /* xi, ai, bi */
     pollard_rho_cycle(U,V,W,p,n,a,g);
     pollard_rho_cycle(U,V,W,p,n,a,g);   /* x2i, a2i, b2i */
     if (verbose > 3) printf( "%3"UVuf"  %4"UVuf" %3"UVuf" %3"UVuf"  %4"UVuf" %3"UVuf" %3"UVuf"\n", i, u, v, w, U, V, W );
     if (u == U) {
-      UV r1, r2, k;
+      UV r1, r2, G, G2;
       r1 = submod(v, V, n);
       if (r1 == 0) {
         if (verbose) printf("DLP Rho failure, r=0\n");
-        return 0;
+        s->failed = 1;
+        k = 0;
+        break;
       }
       r2 = submod(W, w, n);
-      k = divmod(r2, r1, n);
+
+      G = gcd_ui(r1,n);
+      G2 = gcd_ui(G,r2);
+      k = divmod(r2/G2, r1/G2, n/G2);
+      if (G > 1) {
+        if (powmod(g,k,p) == a) {
+          if (verbose > 2) printf("  common GCD %"UVuf"\n", G2);
+        } else {
+          UV m, l = divmod(r2, r1, n/G);
+          for (m = 0; m < G; m++) {
+            k = addmod(l, mulmod(m,(n/G),n), n);
+            if (powmod(g,k,p) == a) break;
+          }
+          if (m<G && verbose > 2) printf("  GCD %"UVuf", found with m=%"UVuf"\n", G, m);
+        }
+      }
+
       if (powmod(g,k,p) != a) {
         if (verbose > 2) printf("r1 = %"UVuf"  r2 = %"UVuf" k = %"UVuf"\n", r1, r2, k);
         if (verbose) printf("Incorrect DLP Rho solution: %"UVuf"\n", k);
-        return 0;
+        s->failed = 1;
+        k = 0;
       }
-      if (verbose) printf("DLP Rho solution found after %"UVuf" steps\n", i);
-      return k;
+      break;
     }
   }
+  s->round += i-1;
+  if (verbose && k) printf("DLP Rho solution found after %"UVuf" steps\n", s->round + 1);
+  s->u = u; s->v = v; s->w = w; s->U = U; s->V = V; s->W = W;
+  return k;
+}
+
+#if 0
+static UV dlp_prho(UV a, UV g, UV p, UV n, UV maxrounds) {
+#ifdef DEBUG
+  int const verbose = _XS_get_verbose()
+#else
+  int const verbose = 0;
+#endif
+  prho_state_t s = {1, 0, 0, 1, 0, 0,   0, 0, verbose};
+  return dlp_prho_uvw(a, g, p, n, maxrounds, &s);
+}
+#endif
+
+
+/******************************************************************************/
+/* DLP - BSGS */
+/******************************************************************************/
+
+typedef struct bsgs_hash_t {
+  UV M;    /* The baby step index */
+  UV V;    /* The powmod value */
+  struct bsgs_hash_t* next;
+} bsgs_hash_t;
+
+/****************************************/
+/*  Simple and limited pool allocation  */
+#define BSGS_ENTRIES_PER_PAGE 8000
+typedef struct bsgs_page_top_t {
+  struct bsgs_page_t* first;
+  bsgs_hash_t** table;
+  UV  size;
+  int nused;
+  int npages;
+} bsgs_page_top_t;
+
+typedef struct bsgs_page_t {
+  bsgs_hash_t entries[BSGS_ENTRIES_PER_PAGE];
+  struct bsgs_page_t* next;
+} bsgs_page_t;
+
+static bsgs_hash_t* get_entry(bsgs_page_top_t* top) {
+  if (top->nused == 0 || top->nused >= BSGS_ENTRIES_PER_PAGE) {
+    bsgs_page_t* newpage;
+    Newz(0, newpage, 1, bsgs_page_t);
+    newpage->next = top->first;
+    top->first = newpage;
+    top->nused = 0;
+    top->npages++;
+  }
+  return top->first->entries + top->nused++;
+}
+static void destroy_pages(bsgs_page_top_t* top) {
+  bsgs_page_t* head = top->first;
+  while (head != 0) {
+    bsgs_page_t* next = head->next;
+    Safefree(head);
+    head = next;
+  }
+  top->first = 0;
+}
+/****************************************/
+
+static void bsgs_hash_put(bsgs_page_top_t* pagetop, UV v, UV i) {
+  UV idx = v % pagetop->size;
+  bsgs_hash_t** table = pagetop->table;
+  bsgs_hash_t* entry = table[idx];
+
+  while (entry && entry->V != v)
+    entry = entry->next;
+
+  if (!entry) {
+    entry = get_entry(pagetop);
+    entry->M = i;
+    entry->V = v;
+    entry->next = table[idx];
+    table[idx] = entry;
+  }
+}
+
+static UV bsgs_hash_get(bsgs_page_top_t* pagetop, UV v) {
+  bsgs_hash_t* entry = pagetop->table[v % pagetop->size];
+  while (entry && entry->V != v)
+    entry = entry->next;
+  return (entry) ? entry->M : 0;
+}
+
+static UV bsgs_hash_put_get(bsgs_page_top_t* pagetop, UV v, UV i) {
+  UV idx = v % pagetop->size;
+  bsgs_hash_t** table = pagetop->table;
+  bsgs_hash_t* entry = table[idx];
+
+  while (entry && entry->V != v)
+    entry = entry->next;
+
+  if (entry)
+    return entry->M;
+
+  entry = get_entry(pagetop);
+  entry->M = i;
+  entry->V = v;
+  entry->next = table[idx];
+  table[idx] = entry;
   return 0;
 }
+
+static UV dlp_bsgs(UV a, UV g, UV p, UV n, UV maxent, int race_rho) {
+  bsgs_page_top_t PAGES;
+  UV i, m, maxm, hashmap_count;
+  UV aa, S, gm, T, gs_i, bs_i;
+  UV result = 0;
+#ifdef DEBUG
+  int const verbose = _XS_get_verbose();
+#else
+  int const verbose = 0;
+#endif
+  prho_state_t rho_state = {1, 0, 0, 1, 0, 0,   0, 0, verbose};
+
+  if (n <= 2) return 0;   /* Shouldn't be here with gorder this low */
+
+  if (race_rho) {
+    result = dlp_prho_uvw(a, g, p, n, 10000, &rho_state);
+    if (result) {
+      if (verbose) printf("rho found solution in BSGS step 0\n");
+      return result;
+    }
+  }
+
+  if (a == 0) return 0;  /* We don't handle this case */
+
+  maxm = isqrt(n);
+  m = (maxent > maxm) ? maxm : maxent;
+
+  hashmap_count = (m < 65537) ? 65537 :
+                  (m > 40000000) ? 40000003 :
+                  next_prime(m);               /* Ave depth around 2 */
+
+  /* Create table.  Size: 8*hashmap_count bytes. */
+  PAGES.size = hashmap_count;
+  PAGES.first = 0;
+  PAGES.nused = 0;
+  PAGES.npages = 0;
+  Newz(0, PAGES.table, hashmap_count, bsgs_hash_t*);
+
+  aa = mulmod(a,a,p);
+  S = a;
+  gm = powmod(g, m, p);
+  T = gm;
+  gs_i = 0;
+  bs_i = 0;
+
+  bsgs_hash_put(&PAGES, S, 0);   /* First baby step */
+  S = mulmod(S, g, p);
+  /* Interleaved Baby Step Giant Step */
+  for (i = 1; i <= m; i++) {
+    gs_i = bsgs_hash_put_get(&PAGES, S, i);
+    if (gs_i) { bs_i = i; break; }
+    S = mulmod(S, g, p);
+    if (S == aa) {  /* We discovered the solution! */
+      if (verbose) printf("  dlp bsgs: solution at BS step %"UVuf"\n", i+1);
+      result = i+1;
+      break;
+    }
+    bs_i = bsgs_hash_put_get(&PAGES, T, i);
+    if (bs_i) { gs_i = i; break; }
+    T = mulmod(T, gm, p);
+    if (race_rho && (i % 2048) == 0) {
+      result = dlp_prho_uvw(a, g, p, n, 100000, &rho_state);
+      if (result) {
+        if (verbose) printf("rho found solution in BSGS step %"UVuf"\n", i);
+        break;
+      }
+    }
+  }
+
+  if (!result) {
+    /* Extend Giant Step search */
+    if (!(gs_i || bs_i)) {
+      UV b = (p+m-1)/m;
+      if (m < maxm && b > 8*m) b = 8*m;
+      for (i = m+1; i < b; i++) {
+        bs_i = bsgs_hash_get(&PAGES, T);
+        if (bs_i) { gs_i = i; break; }
+        T = mulmod(T, gm, p);
+        if (race_rho && (i % 2048) == 0) {
+          result = dlp_prho_uvw(a, g, p, n, 100000, &rho_state);
+          if (result) {
+            if (verbose) printf("rho found solution in BSGS step %"UVuf"\n", i);
+            break;
+          }
+        }
+      }
+    }
+
+    if (gs_i || bs_i) {
+      result = submod(mulmod(gs_i, m, p), bs_i, p);
+    }
+  }
+  if (verbose) printf("  dlp bsgs using %d pages (%.1fMB+%.1fMB) for hash\n", PAGES.npages, ((double)PAGES.npages * sizeof(bsgs_page_t)) / (1024*1024), ((double)hashmap_count * sizeof(bsgs_hash_t*)) / (1024*1024));
+
+  destroy_pages(&PAGES);
+  Safefree(PAGES.table);
+  if (result != 0 && powmod(g,result,p) != a) {
+    if (verbose) printf("Incorrect DLP BSGS solution: %"UVuf"\n", result);
+    result = 0;
+  }
+  if (race_rho && result == 0) {
+    result = dlp_prho_uvw(a, g, p, n, 2000000000U, &rho_state);
+  }
+  return result;
+}
+
+/* Find smallest k where a = g^k mod p */
+#define DLP_TRIAL_NUM  10000
+static UV znlog_solve(UV a, UV g, UV p, UV n) {
+  UV k, sqrtn;
+  const int verbose = _XS_get_verbose();
+
+  if (a >= p) a %= p;
+  if (g >= p) g %= p;
+
+  if (a == 1 || g == 0 || p <= 2)
+    return 0;
+
+  if (verbose > 1 && n != p-1) printf("  g=%"UVuf" p=%"UVuf", order %"UVuf"\n", g, p, n);
+
+  /* printf(" solving znlog(%"UVuf",%"UVuf",%"UVuf") n=%"UVuf"\n", a, g, p, n); */
+
+  if (n == 0 || n <= DLP_TRIAL_NUM) {
+    k = dlp_trial(a, g, p, DLP_TRIAL_NUM);
+    if (verbose) printf("  dlp trial 10k %s\n", (k!=0 || p <= DLP_TRIAL_NUM) ? "success" : "failure");
+    if (k != 0 || (n > 0 && n <= DLP_TRIAL_NUM)) return k;
+  }
+
+  { /* Existence checks */
+    UV aorder, gorder = n;
+    if (gorder != 0 && powmod(a, gorder, p) != 1) return 0;
+    aorder = znorder(a,p);
+    if (aorder == 0 && gorder != 0) return 0;
+    if (aorder != 0 && gorder % aorder != 0) return 0;
+  }
+
+  sqrtn = (n == 0) ? 0 : isqrt(n);
+  if (n == 0) n = p-1;
+
+  {
+    UV maxent = (sqrtn > 0) ? sqrtn+1 : 100000;
+    k = dlp_bsgs(a, g, p, n, maxent/2, /* race rho */ 1);
+    if (verbose) printf("  dlp bsgs %"UVuf"k %s\n", maxent/1000, k!=0 ? "success" : "failure");
+    if (k != 0) return k;
+    if (sqrtn > 0 && sqrtn < maxent) return 0;
+  }
+
+  if (verbose) printf("  dlp doing exhaustive trial\n");
+  k = dlp_trial(a, g, p, p);
+  return k;
+}
+
+/* Silver-Pohlig-Hellman */
+static UV znlog_ph(UV a, UV g, UV p, UV p1) {
+  UV fac[MPU_MAX_FACTORS+1];
+  UV exp[MPU_MAX_FACTORS+1];
+  int i, nfactors;
+  UV x, j;
+
+  if (p1 == 0) return 0;   /* TODO: Should we plow on with p1=p-1? */
+  nfactors = factor_exp(p1, fac, exp);
+  if (nfactors == 1)
+    return znlog_solve(a, g, p, p1);
+  for (i = 0; i < nfactors; i++) {
+    UV pi, delta, gamma;
+    pi = fac[i];   for (j = 1; j < exp[i]; j++)  pi *= fac[i];
+    delta = powmod(a,p1/pi,p);
+    gamma = powmod(g,p1/pi,p);
+    /* printf(" solving znlog(%"UVuf",%"UVuf",%"UVuf")\n", delta, gamma, p); */
+    fac[i] = znlog_solve( delta, gamma, p, znorder(gamma,p) );
+    exp[i] = pi;
+  }
+  x = chinese(fac, exp, nfactors, &i);
+  if (i == 1 && powmod(g, x, p) == a)
+    return x;
+  return 0;
+}
+
+/* Find smallest k where a = g^k mod p */
+UV znlog(UV a, UV g, UV p) {
+  UV k, gorder, aorder;
+  const int verbose = _XS_get_verbose();
+
+  if (a >= p) a %= p;
+  if (g >= p) g %= p;
+
+  if (a == 1 || g == 0 || p <= 2)
+    return 0;
+
+  /* TODO: We call znorder with the same p many times.  We should have a
+   * method for znorder given {phi,nfactors,fac,exp} */
+
+  gorder = znorder(g,p);
+  if (gorder != 0 && powmod(a, gorder, p) != 1) return 0;
+  aorder = znorder(a,p);
+  if (aorder == 0 && gorder != 0) return 0;
+  if (aorder != 0 && gorder % aorder != 0) return 0;
+
+  /* TODO: Come up with a better solution for a=0 */
+  if (a == 0 || p < DLP_TRIAL_NUM || (gorder > 0 && gorder < DLP_TRIAL_NUM)) {
+    if (verbose > 1) printf("  dlp trial znlog(%"UVuf",%"UVuf",%"UVuf")\n",a,g,p);
+    k = dlp_trial(a, g, p, p);
+    return k;
+  }
+
+  if (!is_prob_prime(gorder)) {
+    k = znlog_ph(a, g, p, gorder);
+    if (verbose) printf("  dlp PH %s\n", k!=0 ? "success" : "failure");
+    if (k != 0) return k;
+  }
+
+  return znlog_solve(a, g, p, gorder);
+}
+
+
+/* Compile with:
+ *  gcc -O3 -fomit-frame-pointer -march=native -Wall -DFACTOR_STANDALONE -DSTANDALONE factor.c util.c sieve.c cache.c primality.c lmo.c -lm
+ */
+#ifdef FACTOR_STANDALONE
+#include <errno.h>
+int main(int argc, char *argv[])
+{
+  UV n;
+  UV factors[MPU_MAX_FACTORS+1];
+  int nfactors, i, a;
+
+  if (argc <= 1) { printf("usage: %s  <n>\n", argv[0]); return(1); }
+
+  for (a = 1; a < argc; a++) {
+    n = strtoul(argv[a], 0, 10);
+    if (n == ULONG_MAX && errno == ERANGE) { printf("Argument larger than ULONG_MAX\n"); return(-1); }
+    nfactors = factor(n, factors);
+    printf("%"UVuf":", n);
+    for (i = 0; i < nfactors; i++)
+      printf(" %"UVuf"", factors[i]);
+    printf("\n");
+  }
+
+  return(0);
+}
+#endif
