@@ -48,6 +48,12 @@ sub jordan_totient {
   _validate_positive_integer($n);
   return Math::Prime::Util::PP::jordan_totient($k, $n);
 }
+sub ramanujan_sum {
+  my($k,$n) = @_;
+  _validate_positive_integer($k);
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::ramanujan_sum($k, $n);
+}
 sub carmichael_lambda {
   my($n) = @_;
   _validate_positive_integer($n);
@@ -69,8 +75,24 @@ sub exp_mangoldt {
   _validate_positive_integer($n);
   return Math::Prime::Util::PP::exp_mangoldt($n);
 }
+sub hclassno {
+  my($n) = @_;
+  return 0 if defined $n && int($n) < 0;
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::hclassno($n);
+}
 
 
+sub next_prime {
+  my($n) = @_;
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::next_prime($n);
+}
+sub prev_prime {
+  my($n) = @_;
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::prev_prime($n);
+}
 sub nth_prime {
   my($n) = @_;
   _validate_positive_integer($n);
@@ -108,15 +130,28 @@ sub prime_count_approx {
 }
 sub twin_prime_count {
   my($low,$high) = @_;
-  if (scalar @_ > 1) {
-    _validate_positive_integer($low);
-    _validate_positive_integer($high);
-  } else {
-    ($low,$high) = (2, $low);
-    _validate_positive_integer($high);
-  }
   return Math::Prime::Util::PP::twin_prime_count($low,$high);
 }
+sub ramanujan_prime_count {
+  my($low,$high) = @_;
+  return Math::Prime::Util::PP::ramanujan_prime_count($low,$high);
+}
+sub sum_primes {
+  my($low,$high) = @_;
+  return Math::Prime::Util::PP::sum_primes($low,$high);
+}
+sub print_primes {
+  my($low,$high,$fd) = @_;
+  return Math::Prime::Util::PP::print_primes($low,$high,$fd);
+}
+sub sieve_range {
+  my($n,$width,$depth) = @_;
+  return Math::Prime::Util::PP::sieve_range($n,$width,$depth);
+}
+sub sieve_prime_cluster {
+  my($low,$high,@cl) = @_;
+  return Math::Prime::Util::PP::sieve_prime_cluster($low,$high,@cl);
+}
 sub twin_prime_count_approx {
   my($n) = @_;
   _validate_positive_integer($n);
@@ -132,18 +167,31 @@ sub nth_twin_prime_approx {
   _validate_positive_integer($n);
   return Math::Prime::Util::PP::nth_twin_prime_approx($n);
 }
+sub nth_ramanujan_prime {
+  my($n) = @_;
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::nth_ramanujan_prime($n);
+}
 
 
-*is_prime       = \&Math::Prime::Util::PP::is_prime;
-*is_prob_prime  = \&Math::Prime::Util::PP::is_prob_prime;
-*is_bpsw_prime  = \&Math::Prime::Util::PP::is_bpsw_prime;
+*is_prime          = \&Math::Prime::Util::PP::is_prime;
+*is_prob_prime     = \&Math::Prime::Util::PP::is_prob_prime;
+*is_provable_prime = \&Math::Prime::Util::PP::is_provable_prime;
+*is_bpsw_prime     = \&Math::Prime::Util::PP::is_bpsw_prime;
 
 sub is_pseudoprime {
-  my($n, $base) = @_;
+  my($n, @bases) = @_;
   return 0 if defined $n && int($n) < 0;
   _validate_positive_integer($n);
-  _validate_positive_integer($base) if defined $base;
-  return Math::Prime::Util::PP::is_pseudoprime($n, $base);
+  croak "No bases given to is_strong_pseudoprime" unless @bases;
+  return Math::Prime::Util::PP::is_pseudoprime($n, @bases);
+}
+sub is_euler_pseudoprime {
+  my($n, @bases) = @_;
+  return 0 if defined $n && int($n) < 0;
+  _validate_positive_integer($n);
+  croak "No bases given to is_euler_pseudoprime" unless @bases;
+  return Math::Prime::Util::PP::is_euler_pseudoprime($n, @bases);
 }
 sub is_strong_pseudoprime {
   my($n, @bases) = @_;
@@ -152,6 +200,12 @@ sub is_strong_pseudoprime {
   croak "No bases given to is_strong_pseudoprime" unless @bases;
   return Math::Prime::Util::PP::is_strong_pseudoprime($n, @bases);
 }
+sub is_euler_plumb_pseudoprime {
+  my($n) = @_;
+  return 0 if defined $n && int($n) < 0;
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::is_euler_plumb_pseudoprime($n);
+}
 sub is_lucas_pseudoprime {
   my($n) = @_;
   return 0 if defined $n && int($n) < 0;
@@ -179,10 +233,18 @@ sub is_almost_extra_strong_lucas_pseudoprime {
   return Math::Prime::Util::PP::is_almost_extra_strong_lucas_pseudoprime($n, $increment);
 }
 sub is_perrin_pseudoprime {
+  my($n,$restrict) = @_;
+  return 0 if defined $n && int($n) < 0;
+  $restrict = 0 unless defined $restrict;
+  _validate_positive_integer($n);
+  _validate_positive_integer($restrict);
+  return Math::Prime::Util::PP::is_perrin_pseudoprime($n, $restrict);
+}
+sub is_catalan_pseudoprime {
   my($n) = @_;
   return 0 if defined $n && int($n) < 0;
   _validate_positive_integer($n);
-  return Math::Prime::Util::PP::is_perrin_pseudoprime($n);
+  return Math::Prime::Util::PP::is_catalan_pseudoprime($n);
 }
 sub is_frobenius_pseudoprime {
   my($n, $P, $Q) = @_;
@@ -197,14 +259,86 @@ sub is_frobenius_underwood_pseudoprime {
   _validate_positive_integer($n);
   return Math::Prime::Util::PP::is_frobenius_underwood_pseudoprime($n);
 }
+sub is_frobenius_khashin_pseudoprime {
+  my($n) = @_;
+  return 0 if defined $n && int($n) < 0;
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::is_frobenius_khashin_pseudoprime($n);
+}
 sub is_aks_prime {
   my($n) = @_;
   return 0 if defined $n && int($n) < 0;
   _validate_positive_integer($n);
   return Math::Prime::Util::PP::is_aks_prime($n);
 }
+sub is_ramanujan_prime {
+  my($n) = @_;
+  return 0 if defined $n && int($n) < 0;
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::is_ramanujan_prime($n);
+}
+sub is_mersenne_prime {
+  my($p) = @_;
+  _validate_positive_integer($p);
+  return Math::Prime::Util::PP::is_mersenne_prime($p);
+}
+sub is_square_free {
+  my($n) = @_;
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::is_square_free($n);
+}
+sub is_carmichael {
+  my($n) = @_;
+  #_validate_positive_integer($n);
+  return Math::Prime::Util::PP::is_carmichael($n);
+}
+sub is_quasi_carmichael {
+  my($n) = @_;
+  #_validate_positive_integer($n);
+  return Math::Prime::Util::PP::is_quasi_carmichael($n);
+}
+sub is_primitive_root {
+  my($a,$n) = @_;
+  $n = -$n if defined $n && $n < 0;
+  $a %= $n if defined $a && $a < 0;
+  _validate_positive_integer($a);
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::is_primitive_root($a,$n);
+}
 
 
+sub lucas_sequence {
+  my($n, $P, $Q, $k) = @_;
+  my ($vp, $vq) = ($P, $Q);
+  $vp = -$vp if defined $vp && $vp < 0;
+  $vq = -$vq if defined $vq && $vq < 0;
+  _validate_positive_integer($n);
+  _validate_positive_integer($vp);
+  _validate_positive_integer($vq);
+  _validate_positive_integer($k);
+  return Math::Prime::Util::PP::lucas_sequence(@_);
+}
+sub lucasu {
+  my($P, $Q, $k) = @_;
+  my ($vp, $vq) = ($P, $Q);
+  $vp = -$vp if defined $vp && $vp < 0;
+  $vq = -$vq if defined $vq && $vq < 0;
+  _validate_positive_integer($vp);
+  _validate_positive_integer($vq);
+  _validate_positive_integer($k);
+  return Math::Prime::Util::PP::lucasu(@_);
+}
+sub lucasv {
+  my($P, $Q, $k) = @_;
+  my ($vp, $vq) = ($P, $Q);
+  $vp = -$vp if defined $vp && $vp < 0;
+  $vq = -$vq if defined $vq && $vq < 0;
+  _validate_positive_integer($vp);
+  _validate_positive_integer($vq);
+  _validate_positive_integer($k);
+  return Math::Prime::Util::PP::lucasv(@_);
+}
+
 sub kronecker {
   my($a, $b) = @_;
   my ($va, $vb) = ($a, $b);
@@ -253,7 +387,7 @@ sub znlog {
 
 sub znprimroot {
   my($n) = @_;
-  $n = -$n if defined $n && $n =~ /^-\d+/;   # TODO: fix this for string bigints
+  $n =~ s/^-(\d+)/$1/ if defined $n;
   _validate_positive_integer($n);
   return Math::Prime::Util::PP::znprimroot($n);
 }
@@ -386,6 +520,49 @@ sub invmod {
   _validate_integer($n);
   return Math::Prime::Util::PP::invmod($a,$n);
 }
+sub sqrtmod {
+  my ($a, $n) = @_;
+  _validate_integer($a);
+  _validate_integer($n);
+  return Math::Prime::Util::PP::sqrtmod($a,$n);
+}
+sub addmod {
+  my ($a, $b, $n) = @_;
+  _validate_integer($a); _validate_integer($b>=0?$b:-$b); _validate_integer($n);
+  return Math::Prime::Util::PP::addmod($a,$b, $n);
+}
+sub mulmod {
+  my ($a, $b, $n) = @_;
+  _validate_integer($a); _validate_integer($b>=0?$b:-$b); _validate_integer($n);
+  return Math::Prime::Util::PP::mulmod($a,$b, $n);
+}
+sub divmod {
+  my ($a, $b, $n) = @_;
+  _validate_integer($a); _validate_integer($b>=0?$b:-$b); _validate_integer($n);
+  return Math::Prime::Util::PP::divmod($a,$b, $n);
+}
+sub powmod {
+  my ($a, $b, $n) = @_;
+  _validate_integer($a); _validate_integer($b>=0?$b:-$b); _validate_integer($n);
+  return Math::Prime::Util::PP::powmod($a,$b, $n);
+}
+sub sqrtint {
+  my($n) = @_;
+  _validate_integer($n);
+  return Math::Prime::Util::PP::sqrtint($n);
+}
+sub rootint {
+  my($n, $k, $refp) = @_;
+  _validate_positive_integer($n);
+  _validate_positive_integer($k);
+  return Math::Prime::Util::PP::rootint($n, $k, $refp);
+}
+sub logint {
+  my($n, $b, $refp) = @_;
+  _validate_positive_integer($n);
+  _validate_positive_integer($b);
+  return Math::Prime::Util::PP::logint($n, $b, $refp);
+}
 
 sub legendre_phi {
   my($x, $a) = @_;
@@ -404,13 +581,26 @@ sub chebyshev_psi {
   _validate_positive_integer($n);
   return Math::Prime::Util::PP::chebyshev_psi($n);
 }
+sub ramanujan_tau {
+  my($n) = @_;
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::ramanujan_tau($n);
+}
 
 sub is_power {
-  my($n, $a) = @_;
-  return 0 if defined $n && $n < 0;
-  _validate_positive_integer($n);
+  my($n, $a, $refp) = @_;
+  my $vn = "$n";  $vn =~ s/^-//;
+  _validate_positive_integer($vn);
   _validate_positive_integer($a) if defined $a;
-  return Math::Prime::Util::PP::is_power($n, $a);
+  $vn = '-'.$vn if $n < 0;
+  return Math::Prime::Util::PP::is_power($vn, $a, $refp);
+}
+sub is_prime_power {
+  my($n, $refp) = @_;
+  my $vn = "$n";  $vn =~ s/^-//;
+  _validate_positive_integer($vn);
+  $vn = '-'.$vn if $n < 0;
+  return Math::Prime::Util::PP::is_prime_power($vn, $refp);
 }
 sub valuation {
   my($n, $k) = @_;
@@ -420,6 +610,18 @@ sub valuation {
   _validate_positive_integer($k);
   return Math::Prime::Util::PP::valuation($n, $k);
 }
+sub todigits {
+  my($n,$base,$len) = @_;
+  return Math::Prime::Util::PP::todigits($n,$base,$len);
+}
+sub todigitstring {
+  my($n,$base,$len) = @_;
+  return Math::Prime::Util::PP::todigitstring($n,$base,$len);
+}
+sub fromdigits {
+  my($n,$base) = @_;
+  return Math::Prime::Util::PP::fromdigits($n,$base);
+}
 sub hammingweight {
   my($n) = @_;
   $n = -$n if defined $n && $n < 0;
@@ -507,6 +709,9 @@ sub fordivisors (&$) {    ## no critic qw(ProhibitSubroutinePrototypes)
 sub forpart (&$;$) {    ## no critic qw(ProhibitSubroutinePrototypes)
   Math::Prime::Util::PP::forpart(@_);
 }
+sub forcomp (&$;$) {    ## no critic qw(ProhibitSubroutinePrototypes)
+  Math::Prime::Util::PP::forcomp(@_);
+}
 sub forcomb (&$;$) {    ## no critic qw(ProhibitSubroutinePrototypes)
   Math::Prime::Util::PP::forcomb(@_);
 }
@@ -516,13 +721,75 @@ sub forperm (&$;$) {    ## no critic qw(ProhibitSubroutinePrototypes)
 
 sub vecreduce (&@) {    ## no critic qw(ProhibitSubroutinePrototypes)
   my($sub, @v) = @_;
-  my $count = shift @v;
+
+  # Mastering Perl page 162, works with old Perl
+  my $caller = caller();
+  no strict 'refs'; ## no critic(strict)
+  local(*{$caller.'::a'}) = \my $a;
+  local(*{$caller.'::b'}) = \my $b;
+  $a = shift @v;
   for my $v (@v) {
-    no strict 'refs'; ## no critic(strict)
-    local( ${caller() . '::a'}, ${caller() . '::b'} ) = ($count, $v);
-    $count = $sub->();
+    $b = $v;
+    $a = $sub->();
   }
-  $count;
+  $a;
+}
+
+sub vecany (&@) {       ## no critic qw(ProhibitSubroutinePrototypes)
+  my $sub = shift;
+  { my $pp; local *_ = \$pp;
+    for my $v (@_) { $pp = $v; return 1 if $sub->(); }
+  }
+  undef;
+}
+sub vecall (&@) {       ## no critic qw(ProhibitSubroutinePrototypes)
+  my $sub = shift;
+  { my $pp; local *_ = \$pp;
+    for my $v (@_) { $pp = $v; return if !$sub->(); }
+  }
+  1;
+}
+sub vecnone (&@) {      ## no critic qw(ProhibitSubroutinePrototypes)
+  my $sub = shift;
+  { my $pp; local *_ = \$pp;
+    for my $v (@_) { $pp = $v; return if $sub->(); }
+  }
+  1;
+}
+sub vecnotall (&@) {    ## no critic qw(ProhibitSubroutinePrototypes)
+  my $sub = shift;
+  { my $pp; local *_ = \$pp;
+    for my $v (@_) { $pp = $v; return 1 if !$sub->(); }
+  }
+  undef;
+}
+
+sub vecfirst (&@) {     ## no critic qw(ProhibitSubroutinePrototypes)
+  my $sub = shift;
+  #for (@_) { return $_ if &{$sub}(); }  return undef;
+  { my $pp; local *_ = \$pp;
+    for my $v (@_) { $pp = $v; return $v if $sub->(); }
+  }
+  undef;
+}
+
+sub vecfirstidx (&@) {     ## no critic qw(ProhibitSubroutinePrototypes)
+  my $sub = shift;
+  { my $pp; local *_ = \$pp; my $i = 0;
+    for my $v (@_) { $pp = $v; return $i if $sub->(); $i++; }
+  }
+  -1;
+}
+
+sub vecextract {
+  my($aref, $mask) = @_;
+  croak "vecextract first argument must be an array reference"
+    unless ref($aref) eq 'ARRAY';
+  return Math::Prime::Util::PP::vecextract(@_);
+}
+
+sub sumdigits {
+  return Math::Prime::Util::PP::sumdigits(@_);
 }
 
 1;
@@ -558,7 +825,7 @@ Dana Jacobsen E<lt>dana@acm.orgE<gt>
 
 =head1 COPYRIGHT
 
-Copyright 2014 by Dana Jacobsen E<lt>dana@acm.orgE<gt>
+Copyright 2014-2016 by Dana Jacobsen E<lt>dana@acm.orgE<gt>
 
 This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.
 
