@@ -17,26 +17,33 @@ extern UV  nth_prime(UV x);
 extern UV  nth_prime_upper(UV x);
 extern UV  nth_prime_lower(UV x);
 extern UV  nth_prime_approx(UV x);
-extern UV  prime_count_lower(UV x);
 extern UV  prime_count_upper(UV x);
+extern UV  prime_count_lower(UV x);
 extern UV  prime_count_approx(UV x);
 extern UV  twin_prime_count(UV low, UV high);
 extern UV  twin_prime_count_approx(UV n);
 extern UV  nth_twin_prime(UV n);
 extern UV  nth_twin_prime_approx(UV n);
-extern UV  nth_ramanujan_prime(UV n);
 extern UV* n_ramanujan_primes(UV n);
 extern UV* n_range_ramanujan_primes(UV nlo, UV nhi);
+extern UV* ramanujan_primes(UV* first, UV* last, UV low, UV high);
 extern int is_ramanujan_prime(UV n);
-extern UV ramanujan_prime_count_lower(UV n);
-extern UV ramanujan_prime_count_upper(UV n);
+extern UV  ramanujan_prime_count(UV lo, UV hi);
+extern UV  ramanujan_prime_count_upper(UV n);
+extern UV  ramanujan_prime_count_lower(UV n);
+extern UV  ramanujan_prime_count_approx(UV n);
+extern UV  nth_ramanujan_prime(UV n);
+extern UV  nth_ramanujan_prime_upper(UV n);
+extern UV  nth_ramanujan_prime_lower(UV n);
 extern int sum_primes(UV low, UV high, UV *sum);
 extern void print_primes(UV low, UV high, int fd);
 
 extern int powerof(UV n);
 extern int is_power(UV n, UV a);
 extern UV rootof(UV n, UV k);
+extern int primepower(UV n, UV* prime);
 extern UV valuation(UV n, UV k);
+extern UV logint(UV n, UV b);
 extern UV mpu_popcount(UV n);
 extern UV mpu_popcount_string(const char* ptr, int len);
 
@@ -61,6 +68,8 @@ extern UV binomial(UV n, UV k);
 extern IV gcdext(IV a, IV b, IV* u, IV* v, IV* s, IV* t); /* Ext Euclidean */
 extern UV modinverse(UV a, UV p);              /* Returns 1/a mod p */
 extern UV divmod(UV a, UV b, UV n);            /* Returns a/b mod n */
+extern int sqrtmod(UV* s, UV a, UV p);         /* sqrt(a) mod p */
+extern int sqrtmod_composite(UV* s, UV a,UV n);/* sqrt(a) mod n */
 extern UV chinese(UV* a, UV* n, UV num, int *status); /* Chinese Remainder */
 
 extern UV totient(UV n);
@@ -70,14 +79,26 @@ extern UV carmichael_lambda(UV n);
 extern UV jordan_totient(UV k, UV n);
 extern UV znprimroot(UV n);
 extern UV znorder(UV a, UV n);
-extern int is_primitive_root(UV a, UV p);
+extern int is_primitive_root(UV a, UV n, int nprime);
 #define is_square_free(n)  (moebius(n) != 0)
+extern int is_carmichael(UV n);
+extern UV  is_quasi_carmichael(UV n);  /* Returns number of bases */
 
+extern UV stirling3(UV n, UV m);
 extern IV stirling2(UV n, UV m);
 extern IV stirling1(UV n, UV m);
 
+extern IV hclassno(UV n);
+extern IV ramanujan_tau(UV n);
+
 extern int strnum_minmax(int min, char* a, STRLEN alen, char* b, STRLEN blen);
 
+extern int from_digit_string(UV* n, const char* s, int base);
+extern int from_digit_to_UV(UV* rn, UV* r, int len, int base);
+extern int from_digit_to_str(char** rstr, UV* r, int len, int base);
+extern int to_digit_array(int* bits, UV n, int base, int length);
+extern int to_digit_string(char *s, UV n, int base, int length);
+
 extern int is_catalan_pseudoprime(UV n);
 
 extern UV gcdz(UV x, UV y);
@@ -97,7 +118,7 @@ static UV isqrt(UV n) {
 }
 #endif
 
-#ifdef FUNC_icbrt
+#if defined(FUNC_icbrt) || defined(FUNC_is_perfect_cube)
 static UV icbrt(UV n) {
   UV b, root = 0;
 #if BITS_PER_WORD == 32
@@ -119,6 +140,18 @@ static UV icbrt(UV n) {
 }
 #endif
 
+#if defined(FUNC_ipow)
+static UV ipow(UV n, UV k) {
+  UV p = 1;
+  while (k) {
+    if (k & 1) p *= n;
+    k >>= 1;
+    if (k)     n *= n;
+  }
+  return p;
+}
+#endif
+
 #if defined(FUNC_gcd_ui) || defined(FUNC_lcm_ui)
 /* If we have a very fast ctz, then use the fast FLINT version of gcd */
 #if defined(__GNUC__) && (__GNUC__ >= 4 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
@@ -157,6 +190,49 @@ static int is_perfect_square(UV n)
 }
 #endif
 
+#ifdef FUNC_is_perfect_cube
+static int is_perfect_cube(UV n)
+{
+  UV m;
+  if ((n & 3) == 2) return 0;
+  /* m = n & 511; if ((m*5016427) & (m*95638165) & 438)  return 0; */
+  m = n % 117; if ((m*833230740) & (m*120676722) & 813764715)  return 0;
+  m = n % 133; if ((m*76846229) & (m*305817297) & 306336544)  return 0;
+  m = icbrt(n);
+  return m*m*m == n;
+}
+#endif
+
+#ifdef FUNC_is_perfect_fifth
+static int is_perfect_fifth(UV n)
+{
+  UV m;
+  if ((n & 3) == 2) return 0;
+  m = n %  88; if ((m*85413603) & (m*76260301) & 26476550)  return 0;
+  m = n %  31; if ((m*80682551) & (m*73523539) & 45414528)  return 0;
+  m = n %  41; if ((m*92806493) & (m*130690042) & 35668129)  return 0;
+  /* m = n %  25; if ((m*109794298) & (m*105535723) & 16097553)  return 0; */
+  m = rootof(n, 5);
+  return m*m*m*m*m == n;
+}
+#endif
+
+#ifdef FUNC_is_perfect_seventh
+static int is_perfect_seventh(UV n)
+{
+  UV m;
+  /* if ((n & 3) == 2) return 0; */
+  m = n & 511; if ((m*97259473) & (m*51311663) & 894)  return 0;
+  m = n %  49; if ((m*109645301) & (m*76482737) & 593520192)  return 0;
+  m = n %  71; if ((m*71818386) & (m*38821587) & 35299393)  return 0;
+  /* m = n %  43; if ((m*101368253) & (m*814158665) & 142131408)  return 0; */
+  /* m = n %  29; if ((m*81935611) & (m*84736134) & 37831965)  return 0; */
+  /* m = n % 116; if ((m*348163737) & (m*1539055705) & 2735997248)  return 0; */
+  m = rootof(n, 7);
+  return m*m*m*m*m*m*m == n;
+}
+#endif
+
 #if defined(FUNC_clz) || defined(FUNC_ctz) || defined(FUNC_log2floor)
 /* log2floor(n) gives the location of the first set bit (starting from left)
  * ctz(n)       gives the number of times n is divisible by 2
